--
-- objectAttacher
-- Specialization for attaching objects on a trailer
--
-- @author Geri-G
--
-- edited by PeterJ FS-UK modteam - 16/11/2012
--
--
-- converted to FS17 by PeterJ - 05/10/2018


objectAttacher = {};

function objectAttacher.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Attachable, specializations) and SpecializationUtil.hasSpecialization(AnimatedVehicle, specializations);
end;

function objectAttacher:load(saveGame)

	self.setIsObjectAttached = SpecializationUtil.callSpecializationsFunction("setIsObjectAttached");
	self.searchObjects = objectAttacher.searchObjects;
	self.attachobjects = objectAttacher.attachobjects;
	self.detachobjects = objectAttacher.detachobjects;
	self.OrientJoint = objectAttacher.OrientJoint;
	self.isObjectInRange = objectAttacher.isObjectInRange;
	self.place = {};
	local i=0;
	while true do
		local basename = string.format("vehicle.objectCastPoints.objectCastPoint(%d)",i);
		local place = {};
		place.node1 = Utils.indexToObject(self.components, getXMLString(self.xmlFile,basename.."#frontRight"));
		if place.node1 == nil then
			break;
		end;
		place.node2 = Utils.indexToObject(self.components, getXMLString(self.xmlFile,basename.."#backLeft"));
		place.attacherNode = Utils.indexToObject(self.components, Utils.getNoNil(getXMLString(self.xmlFile,basename.."#attacherNode"),"0>"));
		place.highOffset = Utils.getNoNil(getXMLFloat(self.xmlFile,basename.."#highOffset"),4);
		table.insert(self.place,place);
		i=i+1;
	end;
	self.Attachedobjects = {};
	self.objectsAttached = false;
	self.isObjectAttached = false;
	
	self.automaticLock = Utils.getNoNil(getXMLBool(self.xmlFile,"vehicle.objectCastPoints#automaticLock"),true);
	if self.automaticLock then
		self.animReference = getXMLString(self.xmlFile, "vehicle.objectCastPoints#animRef");
		if self.animReference ~= nil then
			self.animRefTime = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.objectCastPoints#animRefTime"),0.01);
		end;
	end;
	self.lockAnimation = getXMLString(self.xmlFile, "vehicle.objectCastPoints#animName");
end;

function objectAttacher:delete()
end;

function objectAttacher:readStream(streamId, connection)
	local objectAttach = streamReadBool(streamId);
	self:setIsObjectAttached(objectAttach, true);			
end;

function objectAttacher:writeStream(streamId, connection)
    streamWriteBool(streamId, self.isObjectAttached);
end;

function objectAttacher:keyEvent(unicode, sym, modifier, isDown)
end;

function objectAttacher:mouseEvent(posX, posY, isDown, isUp, button)
end;

function objectAttacher:update(dt)
	if self:getIsActive() then
		if not self.automaticLock then
			if self:getIsActiveForInput() then
				if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA3) then
					self:setIsObjectAttached(not self.isObjectAttached);
				end;
			end;
		end;
	end;
end;

function objectAttacher:updateTick(dt)
	if self.automaticLock then
		if self.animReference ~= nil then
			local curAnimRefTime = self:getAnimationTime(self.animReference);
			if curAnimRefTime > self.animRefTime then
				if not self.isObjectAttached then
					self:setIsObjectAttached(true);
				end;
			else
				if self.isObjectAttached then
					self:setIsObjectAttached(false);
				end;
			end;
		end;
	end;
end;

function objectAttacher:draw()
	if self:getIsActive() and not self.automaticLock then
		if self.isObjectAttached then
			g_currentMission:addHelpButtonText(g_i18n:getText("ReleaseLoad"), InputBinding.IMPLEMENT_EXTRA3);
		else
			g_currentMission:addHelpButtonText(g_i18n:getText("LockLoad"), InputBinding.IMPLEMENT_EXTRA3);
		end;
	end;
end;

function objectAttacher:isObjectInRange(place,object)
	local Xmax, Ymax, Zmax = getWorldTranslation(place.node1);
		  Xmax, Ymax, Zmax = worldToLocal(place.attacherNode,Xmax, Ymax, Zmax);
		  
	local Xmin, Ymin, Zmin = getWorldTranslation(place.node2);	
		  Xmin, Ymin, Zmin = worldToLocal(place.attacherNode,Xmin, Ymin, Zmin);
	
	local Xt, Yt, Zt = getWorldTranslation(object);
		  Xt, Yt, Zt = worldToLocal(place.attacherNode,Xt, Yt, Zt);
		
	if (Xt < math.max(Xmax,Xmin) and Xt > math.min(Xmax,Xmin)) and (Zt < math.max(Zmax,Zmin) and Zt > math.min(Zmax,Zmin)) and (Yt<= ((Ymax+Ymin)/2)+place.highOffset and Yt>= (Ymax+Ymin)/2) then
		return true;
	else
		return false;
	end;
end;

function objectAttacher:searchObjects()
	for k,v in pairs(g_currentMission.vehicles) do
		if v ~= self and v ~= self.attacherVehicle then
			local is1stAttached = false;
			for index,components in pairs(v.components) do		
				for k1,v1 in pairs(self.place) do
					local isInRange = self:isObjectInRange(v1,components.node);	
					if isInRange then	
						is1stAttached = self:attachobjects(v1,components.node);
					end;
				end;
			end;
			if is1stAttached then
				self.objectsAttached = true;
			end;
		end;
	end;
end;

function objectAttacher:setIsObjectAttached(objectAttach, noEventSend)
	SetObjectAttachEvent.sendEvent(self, objectAttach, noEventSend)
	self.isObjectAttached = objectAttach;

	if objectAttach then
		self:searchObjects();		
		if self.lockAnimation ~= nil then
			self:playAnimation(self.lockAnimation, 1, nil, true);
		end;
	else
		self:detachobjects();		
		if self.lockAnimation ~= nil then
			self:playAnimation(self.lockAnimation, -1, nil, true);
		end;
	end;
end;

function objectAttacher:OrientJoint(Source, Target)	
	local xw, yw, zw = getWorldTranslation(Source);		
	local x,y,z = worldToLocal(getParent(Target), xw, yw, zw);	
	setTranslation(Target, x,y,z);			
	
	local zX, zY, zZ = localDirectionToWorld(Source, 0,0,1);
	local zX, zY, zZ = worldDirectionToLocal(getParent(Target), zX, zY, zZ);
	local yX, yY, yZ = localDirectionToWorld(Source, 0,1,0);
	local yX, yY, yZ = worldDirectionToLocal(getParent(Target), yX, yY, yZ);
	setDirection(Target, zX, zY, zZ, yX, yY, yZ);
	return false;
end;

function objectAttacher:attachobjects(place,object,objectT)
	local attachedobject = {};
	attachedobject.object = object;
	if self.isServer then
		attachedobject.AT = createTransformGroup("AT");
		link(place.attacherNode,attachedobject.AT);
		
		self:OrientJoint(object,attachedobject.AT);

		local constr = JointConstructor:new();					
		constr:setActors(place.attacherNode, object);
		constr:setJointTransforms(attachedobject.AT,  object);
		for i=1, 3 do
			constr:setTranslationLimit(i-1, true, 0, 0);
			constr:setRotationLimit(i-1,0,0);
		end;
		attachedobject.JointIndex = constr:finalize();
		
		if objectT ~= nil then
			attachedobject.objectT = objectT;
			objectT.isAttached = true;
		end;
	end;

	table.insert(self.Attachedobjects, attachedobject);
	return true;
end;

function objectAttacher:detachobjects()
	if self.isServer then
		for k,v in pairs(self.Attachedobjects) do
			removeJoint(v.JointIndex);
			delete(v.AT);
			v.JointIndex = nil;
			if v.objectT ~= nil then
				v.objectT.isAttached = nil;
			end;
		end;
	end;
	
	self.Attachedobjects = nil;
	self.Attachedobjects = {};
	self.objectsAttached = false;
end;

function objectAttacher:onAttach(attacherVehicle)
	if self.isServer then
		--self:setIsObjectAttached(true);
	end;
end;

function objectAttacher:onDetach()
	if self.isServer then
		if self.isObjectAttached then
			self:setIsObjectAttached(false);
		end;
	end;
end;

SetObjectAttachEvent = {};
SetObjectAttachEvent_mt = Class(SetObjectAttachEvent, Event);
  
InitEventClass(SetObjectAttachEvent, "SetObjectAttachEvent");
  
function SetObjectAttachEvent:emptyNew()
	local self = Event:new(SetObjectAttachEvent_mt);
	self.className="SetObjectAttachEvent";
	return self;
end;
  
function SetObjectAttachEvent:new(object, objectAttach)
	local self = SetObjectAttachEvent:emptyNew()
	self.object = object;
	self.objectAttach = objectAttach;
	return self;
end;
  
function SetObjectAttachEvent:readStream(streamId, connection)
	local id = streamReadInt32(streamId);
	self.objectAttach = streamReadBool(streamId);
	self.object = networkGetObject(id);
	self:run(connection);
end;
  
function SetObjectAttachEvent:writeStream(streamId, connection)
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteBool(streamId, self.objectAttach);
end;
  
function SetObjectAttachEvent:run(connection)
	if not connection:getIsServer() then
		g_server:broadcastEvent(self, false, connection, self.object);
	end;
	self.object:setIsObjectAttached(self.objectAttach, true);
end;
  
function SetObjectAttachEvent.sendEvent(vehicle, objectAttach, noEventSend)
	if objectAttach ~= vehicle.isObjectAttached then
		if noEventSend == nil or noEventSend == false then
			if g_server ~= nil then
				g_server:broadcastEvent(SetObjectAttachEvent:new(vehicle, objectAttach), nil, nil, vehicle);
			else
				g_client:getServerConnection():sendEvent(SetObjectAttachEvent:new(vehicle, objectAttach));
			end;
		end;
	end;
end;